import torch
from .ActivationFunction import ActivationFunction


class TripleStateSwish(ActivationFunction):
    def __init__(self, alpha: float = 20, beta: float = 40, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.alpha = alpha
        self.beta = beta

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        a = 1 / (1 + torch.exp(-x))
        b = 1 / (1 + torch.exp(-x + self.alpha))
        c = 1 / (1 + torch.exp(-x + self.beta))
        return x * a * (a + b + c)



import torch


class HasProbabilities:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def probabilities(self, x: torch.Tensor):
        return torch.softmax(self.logits(x), dim=1)



import torch
from .ActivationFunction import ActivationFunction


class SineReLU(ActivationFunction):
    def __init__(self, epsilon: float = 0.0025):
        super().__init__()
        self.epsilon: float = epsilon

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.where(x >= 0, x, self.epsilon * (torch.sin(x) - torch.cos(x)))



import torch
from .Loss import Loss


class LogitNormLoss(Loss):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, logits, target):
        norms = torch.norm(logits, p=2, dim=-1, keepdim=True) + 1e-7
        normalized_logits = torch.div(logits, norms)
        return torch.nn.functional.cross_entropy(normalized_logits, target)



import torch


class SReLU(torch.nn.Module):
    def __init__(self, channels: int | None = None, max_input: float = 1.0):
        super().__init__()
        self.M = float(max_input)
        if channels is None:
            self.register_parameter("t_l", None)
            self.register_parameter("t_r", None)
            self.register_parameter("a_l", None)
            self.register_parameter("a_r", None)
        else:
            self._init_params(channels, None, None)

    def _init_params(self, C: int, device, dtype):
        self.t_l = torch.nn.Parameter(torch.zeros(C, device=device, dtype=dtype))
        self.t_r = torch.nn.Parameter(
            torch.full((C,), self.M, device=device, dtype=dtype)
        )
        self.a_l = torch.nn.Parameter(torch.zeros(C, device=device, dtype=dtype))
        self.a_r = torch.nn.Parameter(torch.ones(C, device=device, dtype=dtype))

    def _expand_param(p: torch.Tensor, x: torch.Tensor):
        return p.view((1, x.shape[1]) + (1,) * (x.dim() - 2))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.t_l is None:
            self._init_params(x.shape[1], x.device, x.dtype)

        t_l = self._expand_param(self.t_l, x)
        t_r = self._expand_param(self.t_r, x)
        a_l = self._expand_param(self.a_l, x)
        a_r = self._expand_param(self.a_r, x)
        y = torch.where(x < t_l, t_l + a_l * (x - t_l), x)
        y = torch.where(x > t_r, t_r + a_r * (x - t_r), y)
        return y



import torch
from ..AdaptiveActivationFunction import AdaptiveActivationFunction
from .concerns import ChannelBased


class PiLU(AdaptiveActivationFunction, ChannelBased):
    def __init__(self):
        super().__init__()
        self.a = None
        self.b = None
        self.c = None

    def forward(self, x: torch.Tensor):
        self.initialize(x, ["a", "b", "c"], [1, 0.01, 1])
        a = self.a.view(self.parameter_shape(x))
        b = self.b.view(self.parameter_shape(x))
        c = self.c.view(self.parameter_shape(x))
        return torch.where(x >= c, a * x + c * (1 - a), b * x + c * (1 - b))



import torch
import io
from typing import List
from ...vision import Model
from ..utils import get_model_device


class StoresModels:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.models: List[torch.nn.Module] = []

    def record(self, model: Model | torch.nn.Module):
        model_: torch.nn.Module | None = None
        if isinstance(model, Model):
            model_ = model.network
        elif isinstance(model, torch.nn.Module):
            model_ = model
        else:
            raise TypeError("Wrong input to ensemble record")

        device = get_model_device(model_)
        buffer = io.BytesIO()
        torch.save(model_.to("cpu"), buffer)
        buffer.seek(0)
        model_ = torch.load(buffer, map_location=device)
        self.models.append(model_)

    def push(self, *args, **kwargs):
        self.record(*args, **kwargs)

    def append(self, *args, **kwargs):
        self.record(*args, **kwargs)

    def add(self, *args, **kwargs):
        self.record(*args, **kwargs)



from .DualLine import DualLine
from .LeLeLU import LeLeLU
from .AReLU import AReLU
from .PERU import PERU
from .ShiLU import ShiLU
from .StarReLU import StarReLU
from .DPReLU import DPReLU
from .PiLU import PiLU
from .FReLU import FReLU
from .AOAF import AOAF



import torch
from ..AdaptiveActivationFunction import AdaptiveActivationFunction
from ...device import get_device


class AReLU(AdaptiveActivationFunction):
    def __init__(self):
        super(AReLU, self).__init__()
        self.a = torch.nn.Parameter(torch.tensor(0.9, requires_grad=True))
        self.b = torch.nn.Parameter(torch.tensor(2.0, requires_grad=True))
        self.a.to(get_device())
        self.b.to(get_device())

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        negative_slope = torch.clamp(self.a, 0.01, 0.99)
        positive_slope = 1 + torch.sigmoid(self.b)
        positive = positive_slope * torch.relu(x)
        negative = negative_slope * (-torch.relu(-x))
        return positive + negative



import torch
from typing import Sequence, Tuple


class MexicanReLU(torch.nn.Module):
    def __init__(
        self,
        alphas_lambdas: Sequence[Tuple[float, float]],
        channels: int | None = None,
        max_input: float = 1.0,
    ):
        super().__init__()
        self.M = float(max_input)
        self.register_buffer(
            "alphas", torch.tensor([a for a, _ in alphas_lambdas], dtype=torch.float32)
        )
        self.register_buffer(
            "lambdas", torch.tensor([l for _, l in alphas_lambdas], dtype=torch.float32)
        )
        self.K = len(alphas_lambdas)

        if channels is None:
            self.register_parameter("c0", None)  # PReLU negative slope (per-channel)
            self.register_parameter("c", None)  # (C, K) coefficients
        else:
            self._init_params(channels, device=None, dtype=None)

    def _init_params(self, C: int, device, dtype):
        self.c0 = torch.nn.Parameter(torch.zeros(C, device=device, dtype=dtype))
        self.c = torch.nn.Parameter(torch.zeros(C, self.K, device=device, dtype=dtype))

    def _expand_param(p: torch.Tensor, x: torch.Tensor, n_extra: int = 0):
        shape = (
            (1, x.shape[1]) + (1,) * (x.dim() - 2) + ((p.shape[-1],) if n_extra else ())
        )
        return p.view(shape)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if self.c0 is None or self.c is None:
            self._init_params(x.shape[1], x.device, x.dtype)
        c0 = self._expand_param(self.c0, x)
        y = F.relu(x) - c0 * F.relu(-x)
        xE = x.unsqueeze(-1)
        cE = self._expand_param(self.c, x, n_extra=1)
        aE = self.alphas.to(x.device, x.dtype).view(*((1,) * x.dim()), -1)  # (..., K)
        lE = self.lambdas.to(x.device, x.dtype).view(*((1,) * x.dim()), -1)  # (..., K)
        hats = (lE * self.M - (xE - aE * self.M).abs()).clamp_min(0.0)
        y = y + (cE * hats).sum(dim=-1)
        return y



import torch
from .ActivationFunction import ActivationFunction


class LaLU(ActivationFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        phi_laplace = torch.where(x >= 0, 1 - 0.5 * torch.exp(-x), 0.5 * torch.exp(x))
        return x * phi_laplace
